@@ -1,6 +1,8 @@
 import contextlib
 import inspect
 import logging
+import sys
+import types
 import typing
 
 from typing import Any, Callable, Dict, List, Tuple
@@ -16,6 +18,7 @@ from robotpy_ext.autonomous import AutonomousModeSelector
 from robotpy_ext.misc import NotifierDelay
 from robotpy_ext.misc.simple_watchdog import SimpleWatchdog
 
+from .inject import get_injection_requests, find_injections
 from .magic_tunable import setup_tunables, tunable, collect_feedbacks
 from .magic_reset import collect_resets
 
@@ -28,24 +31,24 @@ class MagicInjectError(ValueError):
 
 class MagicRobot(wpilib.RobotBase):
     """
-        Robots that use the MagicBot framework should use this as their
-        base robot class. If you use this as your base, you must
-        implement the following methods:
+    Robots that use the MagicBot framework should use this as their
+    base robot class. If you use this as your base, you must
+    implement the following methods:
 
-        - :meth:`createObjects`
-        - :meth:`teleopPeriodic`
+    - :meth:`createObjects`
+    - :meth:`teleopPeriodic`
 
-        MagicRobot uses the :class:`.AutonomousModeSelector` to allow you
-        to define multiple autonomous modes and to select one of them via
-        the SmartDashboard/Shuffleboard.
+    MagicRobot uses the :class:`.AutonomousModeSelector` to allow you
+    to define multiple autonomous modes and to select one of them via
+    the SmartDashboard/Shuffleboard.
+
+    MagicRobot will set the following NetworkTables variables
+    automatically:
+
+    - ``/robot/mode``: one of 'disabled', 'auto', 'teleop', or 'test'
+    - ``/robot/is_simulation``: True/False
+    - ``/robot/is_ds_attached``: True/False
 
-        MagicRobot will set the following NetworkTables variables
-        automatically:
-        
-        - ``/robot/mode``: one of 'disabled', 'auto', 'teleop', or 'test'
-        - ``/robot/is_simulation``: True/False
-        - ``/robot/is_ds_attached``: True/False
-        
     """
 
     #: Amount of time each loop takes (default is 20ms)
@@ -81,9 +84,14 @@ class MagicRobot(wpilib.RobotBase):
     def _simulationPeriodic(self):
         pass
 
+    def __simulationPeriodic(self):
+        hal.simPeriodicBefore()
+        self._simulationPeriodic()
+        hal.simPeriodicAfter()
+
     def robotInit(self):
         """
-            .. warning:: Internal API, don't override; use :meth:`createObjects` instead
+        .. warning:: Internal API, don't override; use :meth:`createObjects` instead
         """
 
         # Create the user's objects and stuff here
@@ -114,30 +122,30 @@ class MagicRobot(wpilib.RobotBase):
 
         if self.isSimulation():
             self._simulationInit()
-            self.__periodics.append((self._simulationPeriodic, "simulationPeriodic()"))
+            self.__periodics.append((self.__simulationPeriodic, "simulationPeriodic()"))
 
-    def createObjects(self):
+    def createObjects(self) -> None:
         """
-            You should override this and initialize all of your wpilib
-            objects here (and not in your components, for example). This
-            serves two purposes:
+        You should override this and initialize all of your wpilib
+        objects here (and not in your components, for example). This
+        serves two purposes:
 
-            - It puts all of your motor/sensor initialization in the same
-              place, so that if you need to change a port/pin number it
-              makes it really easy to find it. Additionally, if you want
-              to create a simplified robot program to test a specific
-              thing, it makes it really easy to copy/paste it elsewhere
+        - It puts all of your motor/sensor initialization in the same
+          place, so that if you need to change a port/pin number it
+          makes it really easy to find it. Additionally, if you want
+          to create a simplified robot program to test a specific
+          thing, it makes it really easy to copy/paste it elsewhere
 
-            - It allows you to use the magic injection mechanism to share
-              variables between components
+        - It allows you to use the magic injection mechanism to share
+          variables between components
 
-            .. note:: Do not access your magic components in this function,
-                      as their instances have not been created yet. Do not
-                      create them either.
+        .. note:: Do not access your magic components in this function,
+                  as their instances have not been created yet. Do not
+                  create them either.
         """
         raise NotImplementedError
 
-    def autonomousInit(self):
+    def autonomousInit(self) -> None:
         """Initialization code for autonomous mode may go here.
 
         Users may override this method for initialization code which
@@ -146,36 +154,38 @@ class MagicRobot(wpilib.RobotBase):
 
         This can be useful for code that must be run at the beginning of a match.
 
-        .. note:: The ``on_enable`` functions of all components are
-                  called before this function is called.
+        .. note::
+
+           This method is called after every component's ``on_enable`` method,
+           but before the selected autonomous mode's ``on_enable`` method.
         """
         pass
 
-    def teleopInit(self):
+    def teleopInit(self) -> None:
         """
-            Initialization code for teleop control code may go here.
+        Initialization code for teleop control code may go here.
 
-            Users may override this method for initialization code which will be
-            called each time the robot enters teleop mode.
+        Users may override this method for initialization code which will be
+        called each time the robot enters teleop mode.
 
-            .. note:: The ``on_enable`` functions of all components are called
-                      before this function is called.
+        .. note:: The ``on_enable`` functions of all components are called
+                  before this function is called.
         """
         pass
 
     def teleopPeriodic(self):
         """
-            Periodic code for teleop mode should go here.
+        Periodic code for teleop mode should go here.
 
-            Users should override this method for code which will be called
-            periodically at a regular rate while the robot is in teleop mode.
+        Users should override this method for code which will be called
+        periodically at a regular rate while the robot is in teleop mode.
 
-            This code executes before the ``execute`` functions of all
-            components are called.
+        This code executes before the ``execute`` functions of all
+        components are called.
 
-            .. note:: If you want this function to be called in autonomous
-                      mode, set ``use_teleop_in_autonomous`` to True in your
-                      robot class.
+        .. note:: If you want this function to be called in autonomous
+                  mode, set ``use_teleop_in_autonomous`` to True in your
+                  robot class.
         """
         func = self.teleopPeriodic.__func__
         if not hasattr(func, "firstRun"):
@@ -184,27 +194,27 @@ class MagicRobot(wpilib.RobotBase):
             )
             func.firstRun = False
 
-    def disabledInit(self):
+    def disabledInit(self) -> None:
         """
-            Initialization code for disabled mode may go here.
+        Initialization code for disabled mode may go here.
 
-            Users may override this method for initialization code which will be
-            called each time the robot enters disabled mode.
+        Users may override this method for initialization code which will be
+        called each time the robot enters disabled mode.
 
-            .. note:: The ``on_disable`` functions of all components are called
-                      before this function is called.
+        .. note:: The ``on_disable`` functions of all components are called
+                  before this function is called.
         """
         pass
 
     def disabledPeriodic(self):
         """
-            Periodic code for disabled mode should go here.
+        Periodic code for disabled mode should go here.
 
-            Users should override this method for code which will be called
-            periodically at a regular rate while the robot is in disabled mode.
+        Users should override this method for code which will be called
+        periodically at a regular rate while the robot is in disabled mode.
 
-            This code executes before the ``execute`` functions of all
-            components are called.
+        This code executes before the ``execute`` functions of all
+        components are called.
         """
         func = self.disabledPeriodic.__func__
         if not hasattr(func, "firstRun"):
@@ -213,7 +223,7 @@ class MagicRobot(wpilib.RobotBase):
             )
             func.firstRun = False
 
-    def testInit(self):
+    def testInit(self) -> None:
         """Initialization code for test mode should go here.
 
         Users should override this method for initialization code which will be
@@ -221,23 +231,23 @@ class MagicRobot(wpilib.RobotBase):
         """
         pass
 
-    def testPeriodic(self):
+    def testPeriodic(self) -> None:
         """Periodic code for test mode should go here."""
         pass
 
-    def robotPeriodic(self):
+    def robotPeriodic(self) -> None:
         """
-            Periodic code for all modes should go here.
+        Periodic code for all modes should go here.
 
-            Users must override this method to utilize it
-            but it is not required.
+        Users must override this method to utilize it
+        but it is not required.
 
-            This function gets called last in each mode.
-            You may use it for any code you need to run
-            during all modes of the robot (e.g NetworkTables updates)
+        This function gets called last in each mode.
+        You may use it for any code you need to run
+        during all modes of the robot (e.g NetworkTables updates)
 
-            The default implementation will update
-            SmartDashboard, LiveWindow and Shuffleboard.
+        The default implementation will update
+        SmartDashboard, LiveWindow and Shuffleboard.
         """
         watchdog = self.watchdog
         self.__sd_update()
@@ -249,35 +259,35 @@ class MagicRobot(wpilib.RobotBase):
 
     def onException(self, forceReport: bool = False) -> None:
         """
-            This function must *only* be called when an unexpected exception
-            has occurred that would otherwise crash the robot code. Use this
-            inside your :meth:`operatorActions` function.
+        This function must *only* be called when an unexpected exception
+        has occurred that would otherwise crash the robot code. Use this
+        inside your :meth:`operatorActions` function.
 
-            If the FMS is attached (eg, during a real competition match),
-            this function will return without raising an error. However,
-            it will try to report one-off errors to the Driver Station so
-            that it will be recorded in the Driver Station Log Viewer.
-            Repeated errors may not get logged.
+        If the FMS is attached (eg, during a real competition match),
+        this function will return without raising an error. However,
+        it will try to report one-off errors to the Driver Station so
+        that it will be recorded in the Driver Station Log Viewer.
+        Repeated errors may not get logged.
 
-            Example usage::
+        Example usage::
 
-                def teleopPeriodic(self):
-                    try:
-                        if self.joystick.getTrigger():
-                            self.shooter.shoot()
-                    except:
-                        self.onException()
+            def teleopPeriodic(self):
+                try:
+                    if self.joystick.getTrigger():
+                        self.shooter.shoot()
+                except:
+                    self.onException()
 
-                    try:
-                        if self.joystick.getRawButton(2):
-                            self.ball_intake.run()
-                    except:
-                        self.onException()
+                try:
+                    if self.joystick.getRawButton(2):
+                        self.ball_intake.run()
+                except:
+                    self.onException()
 
-                    # and so on...
+                # and so on...
 
-            :param forceReport: Always report the exception to the DS. Don't
-                                set this to True
+        :param forceReport: Always report the exception to the DS. Don't
+                            set this to True
         """
         # If the FMS is not attached, crash the robot program
         if not self.ds.isFMSAttached():
@@ -301,26 +311,26 @@ class MagicRobot(wpilib.RobotBase):
     @contextlib.contextmanager
     def consumeExceptions(self, forceReport: bool = False):
         """
-            This returns a context manager which will consume any uncaught
-            exceptions that might otherwise crash the robot.
+        This returns a context manager which will consume any uncaught
+        exceptions that might otherwise crash the robot.
 
-            Example usage::
+        Example usage::
 
-                def teleopPeriodic(self):
-                    with self.consumeExceptions():
-                        if self.joystick.getTrigger():
-                            self.shooter.shoot()
+            def teleopPeriodic(self):
+                with self.consumeExceptions():
+                    if self.joystick.getTrigger():
+                        self.shooter.shoot()
 
-                    with self.consumeExceptions():
-                        if self.joystick.getRawButton(2):
-                            self.ball_intake.run()
+                with self.consumeExceptions():
+                    if self.joystick.getRawButton(2):
+                        self.ball_intake.run()
 
-                    # and so on...
+                # and so on...
 
-            :param forceReport: Always report the exception to the DS. Don't
-                                set this to True
+        :param forceReport: Always report the exception to the DS. Don't
+                            set this to True
 
-            .. seealso:: :meth:`onException` for more details
+        .. seealso:: :meth:`onException` for more details
         """
         try:
             yield
@@ -359,12 +369,12 @@ class MagicRobot(wpilib.RobotBase):
     def endCompetition(self) -> None:
         self.__done = True
 
-    def autonomous(self):
+    def autonomous(self) -> None:
         """
-            MagicRobot will do The Right Thing and automatically load all
-            autonomous mode routines defined in the autonomous folder.
+        MagicRobot will do The Right Thing and automatically load all
+        autonomous mode routines defined in the autonomous folder.
 
-            .. warning:: Internal API, don't override
+        .. warning:: Internal API, don't override
         """
 
         self.__nt_put_mode("auto")
@@ -377,9 +387,10 @@ class MagicRobot(wpilib.RobotBase):
         except:
             self.onException(forceReport=True)
 
-        auto_functions = (self._execute_components, self._update_feedback,) + tuple(
-            p[0] for p in self.__periodics
-        )
+        auto_functions: Tuple[Callable[[], None], ...] = (
+            self._execute_components,
+            self._update_feedback,
+        ) + tuple(p[0] for p in self.__periodics)
 
         if self.use_teleop_in_autonomous:
             auto_functions = (self.teleopPeriodic,) + auto_functions
@@ -393,13 +404,13 @@ class MagicRobot(wpilib.RobotBase):
 
         self._on_mode_disable_components()
 
-    def _disabled(self):
+    def _disabled(self) -> None:
         """
-            This function is called in disabled mode. You should not
-            override this function; rather, you should override the
-            :meth:`disabledPeriodic` function instead.
+        This function is called in disabled mode. You should not
+        override this function; rather, you should override the
+        :meth:`disabledPeriodic` function instead.
 
-            .. warning:: Internal API, don't override
+        .. warning:: Internal API, don't override
         """
         watchdog = self.watchdog
         watchdog.reset()
@@ -438,13 +449,13 @@ class MagicRobot(wpilib.RobotBase):
                 delay.wait()
                 watchdog.reset()
 
-    def _operatorControl(self):
+    def _operatorControl(self) -> None:
         """
-            This function is called in teleoperated mode. You should not
-            override this function; rather, you should override the
-            :meth:`teleopPeriodics` function instead.
+        This function is called in teleoperated mode. You should not
+        override this function; rather, you should override the
+        :meth:`teleopPeriodics` function instead.
 
-            .. warning:: Internal API, don't override
+        .. warning:: Internal API, don't override
         """
         watchdog = self.watchdog
         watchdog.reset()
@@ -489,7 +500,7 @@ class MagicRobot(wpilib.RobotBase):
 
         self._on_mode_disable_components()
 
-    def _test(self):
+    def _test(self) -> None:
         """Called when the robot is in test mode"""
         watchdog = self.watchdog
         watchdog.reset()
@@ -569,12 +580,7 @@ class MagicRobot(wpilib.RobotBase):
 
         # - Iterate over class variables with type annotations
         # .. this hack is necessary for pybind11 based modules
-        class FakeModule:
-            pass
-
-        import sys
-
-        sys.modules["pybind11_builtins"] = FakeModule()
+        sys.modules["pybind11_builtins"] = types.SimpleNamespace()
 
         for m, ctyp in typing.get_type_hints(cls).items():
             # Ignore private variables
@@ -597,25 +603,7 @@ class MagicRobot(wpilib.RobotBase):
             # Store for later
             components.append((m, component))
 
-        # Collect attributes of this robot that are injectable
-        self._injectables = {}
-
-        for n in dir(self):
-            if (
-                n.startswith("_")
-                or n in self._exclude_from_injection
-                or isinstance(getattr(cls, n, None), tunable)
-            ):
-                continue
-
-            o = getattr(self, n)
-
-            # Don't inject methods
-            # TODO: This could actually be a cool capability..
-            if inspect.ismethod(o):
-                continue
-
-            self._injectables[n] = o
+        self._injectables = self._collect_injectables()
 
         # For each new component, perform magic injection
         for cname, component in components:
@@ -648,6 +636,29 @@ class MagicRobot(wpilib.RobotBase):
 
         self._components = components
 
+    def _collect_injectables(self) -> Dict[str, Any]:
+        injectables = {}
+        cls = type(self)
+
+        for n in dir(self):
+            if (
+                n.startswith("_")
+                or n in self._exclude_from_injection
+                or isinstance(getattr(cls, n, None), (property, tunable))
+            ):
+                continue
+
+            o = getattr(self, n)
+
+            # Don't inject methods
+            # TODO: This could actually be a cool capability..
+            if inspect.ismethod(o):
+                continue
+
+            injectables[n] = o
+
+        return injectables
+
     def _create_component(self, name: str, ctyp: type):
         # Create instance, set it on self
         component = ctyp()
@@ -667,70 +678,22 @@ class MagicRobot(wpilib.RobotBase):
 
         return component
 
-    def _setup_vars(self, cname: str, component):
-
+    def _setup_vars(self, cname: str, component) -> None:
         self.logger.debug("Injecting magic variables into %s", cname)
 
-        component_type = type(component)
+        type_hints = typing.get_type_hints(type(component))
+        requests = get_injection_requests(type_hints, cname, component)
+        injections = find_injections(requests, self._injectables, cname)
+        component.__dict__.update(injections)
 
-        # Iterate over variables with type annotations
-        for n, inject_type in typing.get_type_hints(component_type).items():
-
-            # If the variable is private ignore it
-            if n.startswith("_"):
-                continue
-
-            # If the variable has been set, skip it
-            if hasattr(component, n):
-                continue
-
-            # Check for generic types from the typing module
-            origin = getattr(inject_type, "__origin__", None)
-            if origin is not None:
-                inject_type = origin
-
-            # If the type is not actually a type, give a meaningful error
-            if not isinstance(inject_type, type):
-                raise TypeError(
-                    "Component %s has a non-type annotation on %s (%r); lone non-injection variable annotations are disallowed, did you want to assign a static variable?"
-                    % (cname, n, inject_type)
-                )
-
-            self._inject(n, inject_type, cname, component)
-
-    def _inject(self, n: str, inject_type: type, cname: str, component):
-        # Retrieve injectable object
-        injectable = self._injectables.get(n)
-        if injectable is None:
-            if cname is not None:
-                # Try for mangled names
-                injectable = self._injectables.get("%s_%s" % (cname, n))
-
-        # Raise error if injectable syntax used but no injectable was found.
-        if injectable is None:
-            raise MagicInjectError(
-                "Component %s has variable %s (type %s), which is not present in %s"
-                % (cname, n, inject_type, self)
-            )
-
-        # Raise error if injectable declared with type different than the initial type
-        if not isinstance(injectable, inject_type):
-            raise MagicInjectError(
-                "Component %s variable %s in %s are not the same types! (Got %s, expected %s)"
-                % (cname, n, self, type(injectable), inject_type)
-            )
-        # Perform injection
-        setattr(component, n, injectable)
-        self.logger.debug("-> %s as %s.%s", injectable, cname, n)
-
-    def _setup_reset_vars(self, component):
+    def _setup_reset_vars(self, component) -> None:
         reset_dict = collect_resets(type(component))
 
         if reset_dict:
             component.__dict__.update(reset_dict)
             self._reset_components.append((reset_dict, component))
 
-    def _update_feedback(self):
+    def _update_feedback(self) -> None:
         for method, entry in self._feedbacks:
             try:
                 value = method()
@@ -740,7 +703,7 @@ class MagicRobot(wpilib.RobotBase):
             entry.setValue(value)
         self.watchdog.addEpoch("@magicbot.feedback")
 
-    def _execute_components(self):
+    def _execute_components(self) -> None:
         for name, component in self._components:
             try:
                 component.execute()
